/*
 *
 *    ******************************************************************************
 *    *
 *    * This program and the accompanying materials are made available under the
 *    * terms of the Apache License, Version 2.0 which is available at
 *    * https://www.apache.org/licenses/LICENSE-2.0.
 *    *
 *    *  See the NOTICE file distributed with this work for additional
 *    *  information regarding copyright ownership.
 *    * Unless required by applicable law or agreed to in writing, software
 *    * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *    * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *    * License for the specific language governing permissions and limitations
 *    * under the License.
 *    *
 *    * SPDX-License-Identifier: Apache-2.0
 *    *****************************************************************************
 *
 */

package net.brutex.ai.dnn.api;

import java.util.Collection;
import java.util.Map;
import lombok.NonNull;
import org.deeplearning4j.nn.api.ITrainableLayer;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.gradient.Gradient;
import org.deeplearning4j.nn.workspace.LayerWorkspaceMgr;
import org.deeplearning4j.optimize.api.ConvexOptimizer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.nd4j.common.primitives.Pair;
import org.nd4j.evaluation.IEvaluation;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.api.DataSet;
import org.nd4j.linalg.dataset.api.MultiDataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.dataset.api.iterator.MultiDataSetIterator;

/**
 * A Neural Network is an instance of a {@link INeuralNetworkConfiguration}, that can be trained,
 * evaluated, saved, exported, etc. Its configuration state is defined with the
 * {@link #setNetConfiguration(NeuralNetConfiguration)} (INeuralNetworkConfiguration)} and
 * {@link #getNetConfiguration()} methods.
 **/

public interface IModel extends ITrainableLayer {

  /**
   * The full param table for the model. Each layer may get a subset of its parameters.
   *
   * @return full table of parameters
   */
  Map<String, INDArray> getParamTable();

  Map<String, INDArray> getParamTable(boolean backpropOnly);
  void setParamTable(Map<String, INDArray> paramTable);

  /**
   * This method returns updater state (if applicable), null otherwise
   *
   * @return
   */
  INDArray updaterState();

  /**
   * This method returns Optimizer used for training
   *
   * @return
   */
  ConvexOptimizer getOptimizer();

  /**
   * This method fits model with a given DataSet
   *
   * @param dataSet
   */
  void fit(DataSet dataSet);

  /**
   * This method fits model with a given MultiDataSet
   *
   * @param dataSet
   */
  void fit(MultiDataSet dataSet);

  /**
   * This method fits model with a given DataSetIterator
   *
   * @param iterator
   */
  void fit(DataSetIterator iterator);

  /**
   * This method fits model with a given MultiDataSetIterator
   *
   * @param iterator
   */
  void fit(MultiDataSetIterator iterator);

  /**
   * This method executes evaluation of the model against given iterator and evaluation
   * implementations
   *
   * @param iterator
   */
  <T extends IEvaluation> T[] doEvaluation(DataSetIterator iterator, T... evaluations);

  /**
   * This method executes evaluation of the model against given iterator and evaluation
   * implementations
   *
   * @param iterator
   */
  <T extends IEvaluation> T[] doEvaluation(MultiDataSetIterator iterator, T... evaluations);

  /**
   * Get the configuration of this model.
   *
   * @return the neural net configuration
   */
  NeuralNetConfiguration getNetConfiguration();

  void setNetConfiguration(@NonNull NeuralNetConfiguration netConfiguration);

  /**
   * Init the model
   */
  void init();

  /**
   * Get the number of parameters in this model
   *
   * @return number of parameters
   */
  long numParams();

  /**
   * All models have a fit method
   */
  @Deprecated
  void fit();

  /**
   * Update layer weights and biases with gradient change
   */
  void update(Gradient gradient);

  /**
   * Perform one update applying the gradient
   *
   * @param gradient the gradient to apply
   */
  void update(INDArray gradient, String paramType);


  /**
   * The score for the model. No calculation occurs, this simply returns the score calculated before
   * by the {@link #computeGradientAndScore(LayerWorkspaceMgr)} method.
   *
   * @return the score for the model
   */
  double getScore();


  /**
   * Update the score
   */
  void computeGradientAndScore(LayerWorkspaceMgr workspaceMgr);

  /**
   * Parameters of the model (if any)
   *
   * @return the parameters of the model
   */
  INDArray getModelParams();


  /**
   * the number of parameters for the model
   *
   * @return the number of parameters for the model
   */
  long numParams(boolean backwards);

  /**
   * Set the parameters for this model. This expects a linear ndarray which then be unpacked
   * internally relative to the expected ordering of the model
   *
   * @param params the parameters for the model
   */
  void setParams(INDArray params);

  /**
   * Set the initial parameters array as a view of the full (backprop) network parameters NOTE: this
   * is intended to be used internally in MultiLayerNetwork and ComputationGraph, not by users.
   *
   * @param params a 1 x nParams row vector that is a view of the larger (MLN/CG) parameters array
   */
  void setParamsViewArray(INDArray params);


  INDArray getGradientsViewArray();

  /**
   * Set the gradients array as a view of the full (backprop) network parameters NOTE: this is
   * intended to be used internally in MultiLayerNetwork and ComputationGraph, not by users.
   *
   * @param gradients a 1 x nParams row vector that is a view of the larger (MLN/CG) gradients
   *                  array
   */
  void setBackpropGradientsViewArray(INDArray gradients);

  /**
   * Fit the model to the given data
   *
   * @param data the data to fit the model to
   */
  void fit(INDArray data, LayerWorkspaceMgr workspaceMgr);


  /**
   * Get the gradient. Note that this method will not calculate the gradient, it will rather return
   * the gradient that has been computed before. For calculating the gradient, see
   * {@link IModel#computeGradientAndScore(LayerWorkspaceMgr)} } .
   *
   * @return the gradient for this model, as calculated before
   */
  Gradient gradient();

  /**
   * Get the gradient and score
   *
   * @return the gradient and score
   */
  Pair<Gradient, Double> gradientAndScore();

  /**
   * The current inputs batch size
   *
   * @return the current inputs batch size
   */
  int batchSize();

  /**
   * The input/feature matrix for the model
   *
   * @return the input/feature matrix for the model
   */
  INDArray input();

  /**
   * Get a parameter array for a given parameter type key
   *
   * @param param the key of the parameter
   * @return ndarray of parameters
   */
  INDArray getParam(String param);


  /**
   * Set the parameters for a given parameter type.
   *
   * @param key the param type key to set
   * @param val the new parameters ndarray
   */
  void setParam(String key, INDArray val);

  /**
   * Clear input
   */
  void clear();


  /**
   * Apply any constraints to the model
   */
  void applyConstraints(int iteration, int epoch);


  void close();

  /**
   * Get the TrainingListeners
   *
   * @return training listener
   */
  Collection<TrainingListener> getTrainingListeners();

  /**
   * Replace the TrainingListeners for this model
   *
   * @param listeners new listeners
   */
  void addTrainingListeners(TrainingListener... listeners);

  void addTrainingListeners(Collection<TrainingListener> listeners);


}
