/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.conf;

import com.fasterxml.jackson.annotation.JsonIgnoreProperties;
import com.fasterxml.jackson.databind.JsonNode;
import com.fasterxml.jackson.databind.ObjectMapper;
import com.fasterxml.jackson.databind.node.ArrayNode;
import java.io.IOException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;
import java.util.List;
import java.util.Map;
import java.util.Random;
import lombok.Data;
import lombok.EqualsAndHashCode;
import lombok.Getter;
import lombok.NonNull;
import lombok.Setter;
import lombok.experimental.SuperBuilder;
import lombok.extern.slf4j.Slf4j;
import net.brutex.ai.dnn.api.INeuralNetworkConfiguration;
import org.deeplearning4j.nn.api.OptimizationAlgorithm;
import org.deeplearning4j.nn.api.layers.LayerConstraint;
import org.deeplearning4j.nn.conf.distribution.Distribution;
import org.deeplearning4j.nn.conf.dropout.Dropout;
import org.deeplearning4j.nn.conf.dropout.IDropout;
import org.deeplearning4j.nn.conf.inputs.InputType;
import org.deeplearning4j.nn.conf.layers.BaseLayerConfiguration;
import org.deeplearning4j.nn.conf.layers.ConvolutionLayer;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.conf.serde.JsonMappers;
import org.deeplearning4j.nn.conf.stepfunctions.StepFunction;
import org.deeplearning4j.nn.conf.weightnoise.IWeightNoise;
import org.deeplearning4j.nn.weights.IWeightInit;
import org.deeplearning4j.nn.weights.WeightInit;
import org.deeplearning4j.nn.weights.WeightInitDistribution;
import org.deeplearning4j.nn.weights.WeightInitXavier;
import org.deeplearning4j.util.NetworkUtils;
import org.nd4j.common.base.Preconditions;
import org.nd4j.linalg.activations.Activation;
import org.nd4j.linalg.activations.IActivation;
import org.nd4j.linalg.activations.impl.ActivationSigmoid;
import org.nd4j.linalg.api.buffer.DataType;
import org.nd4j.linalg.learning.config.IUpdater;
import org.nd4j.linalg.learning.config.Sgd;
import org.nd4j.linalg.learning.regularization.L1Regularization;
import org.nd4j.linalg.learning.regularization.L2Regularization;
import org.nd4j.linalg.learning.regularization.Regularization;
import org.nd4j.linalg.learning.regularization.WeightDecay;

/**
 * Deeplearning4j is a domain-specific language to configure deep neural networks, which are made of
 * multiple layers. Everything starts with a NeuralNetConfiguration, which organizes those layers
 * and their hyperparameters. Hyperparameters are variables that determine how a neural network
 * learns. They include how many times to update the weights of the model, how to initialize those
 * weights, which activation function to attach to the nodes, which optimization algorithm to use,
 * and how fast the model should learn. This is what one configuration would look like: <br>
 * <br>
 * NeuralNetConfiguration conf = NeuralNetConfiguration.builder()<br>
 * .weightInit(WeightInit.XAVIER) .activation(Activation.RELU)<br>
 * .optimizationAlgo(OptimizationAlgorithm.STOCHASTIC_GRADIENT_DESCENT)<br>
 * .updater(new Sgd(0.05)) //... other hyperparameters <br>
 * .backprop(true)<br>
 * .build();<br>
 * <br>
 * With Deeplearning4j, you add a layer by calling layer on the
 * NeuralNetConfiguration.NeuralNetConfigurationBuilder(), specifying its place in the order of
 * layers (the zero-indexed layer below is the input layer), the number of input and output nodes,
 * nIn and nOut, as well as the type: DenseLayer.<br>
 * <br>
 * .layer(0, new DenseLayer.Builder().nIn(784).nOut(250)<br>
 * .build())<br>
 * <br>
 * Once you've configured your net, you train the model with model.fit.
 */
@Data
@Slf4j
@EqualsAndHashCode(exclude = {"iterationCount", "epochCount"})
@JsonIgnoreProperties(ignoreUnknown = true)
// The inner builder, that we can then extend ...
@SuperBuilder // TODO fix access
public abstract class NeuralNetBaseBuilderConfiguration implements INeuralNetworkConfiguration {

  private static final int DEFAULT_TBPTT_LENGTH = 20;

  /**
   * Set constraints to be applied to all layers. Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
   * regularization, etc). These constraints are applied at each iteration, after the parameters
   * have been updated.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param constraints Constraints to apply to all weight parameters of all layers
   */
  @lombok.Builder.Default protected final List<LayerConstraint> contrainWeights = new ArrayList<>();

  /**
   * Set constraints to be applied to all layers. Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
   * regularization, etc). These constraints are applied at each iteration, after the parameters
   * have been updated.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param constraints Constraints to apply to all bias parameters of all layers
   */
  @lombok.Builder.Default protected final List<LayerConstraint> biasConstraints = new ArrayList<>();
  /**
   * Set constraints to be applied to all layers. Default: no constraints.<br>
   * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
   * regularization, etc). These constraints are applied at each iteration, after the parameters
   * have been updated.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param constraints Constraints to apply to all parameters of all layers
   */
  @lombok.Builder.Default
  protected final List<LayerConstraint> allParamContraints = new ArrayList<>();
  /**
   * This is a basic concept, a neural network is made of layers, but also can use another neural
   * network as a building block. When the configuration is initialized, those building blocks will
   * be flattened into a single list of layers. Internal ordered list of layers and inner neural
   * networks. If the object is a NeuralNetConfiguration, each configuration must contain at least
   * one layer.
   */
  @Getter @lombok.Builder.Default
  protected final List<Object> innerConfigurations = new ArrayList<>();

  @Getter @Setter @NonNull @lombok.Builder.Default @Deprecated
  protected WorkspaceMode trainingWorkspaceMode = WorkspaceMode.ENABLED;

  @Getter @Setter @NonNull @lombok.Builder.Default @Deprecated
  protected WorkspaceMode inferenceWorkspaceMode = WorkspaceMode.ENABLED;
  /**
   * The type of backprop. Default setting is used for most networks (MLP, CNN etc), but optionally
   * truncated BPTT can be used for training recurrent neural networks. If using TruncatedBPTT make
   * sure you set both tBPTTForwardLength() and tBPTTBackwardLength()
   */
  @Getter @Setter @NonNull @lombok.Builder.Default
  protected BackpropType backpropType = BackpropType.Standard;

  @Getter @lombok.Builder.Default
  protected Map<Integer, InputPreProcessor> inputPreProcessors = new HashMap<>();
  /**
   * When doing truncated BPTT: how many steps of forward pass should we do before doing (truncated)
   * backprop?<br>
   * Only applicable when doing backpropType(BackpropType.TruncatedBPTT)<br>
   * Typically tBPTTForwardLength parameter is same as the tBPTTBackwardLength parameter, but may be
   * larger than it in some circumstances (but never smaller)<br>
   * Ideally your training data time series length should be divisible by this This is the k1
   * parameter on pg23 of <a
   * href="http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf">http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf</a>
   *
   * @param forwardLength Forward length > 0, >= backwardLength
   */
  @Getter @Setter @lombok.Builder.Default protected int tbpttFwdLength = 20;
  /**
   * When doing truncated BPTT: how many steps of backward should we do?<br>
   * Only applicable when doing backpropType(BackpropType.TruncatedBPTT)<br>
   * This is the k2 parameter on pg23 of <a
   * href="http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf">http://www.cs.utoronto.ca/~ilya/pubs/ilya_sutskever_phd_thesis.pdf</a>
   *
   * @param backwardLength <= forwardLength
   */
  @Getter @Setter @lombok.Builder.Default protected int tbpttBackLength = 20;
  // Counter for the number of parameter updates so far
  // This is important for learning rate schedules, for example, and is stored here to ensure it is
  // persisted
  // for Spark and model serialization
  @Getter @Setter @lombok.Builder.Default protected int iterationCount = 0;
  // Counter for the number of epochs completed so far. Used for per-epoch schedules
  @Getter @Setter @lombok.Builder.Default protected int epochCount = 0;
  @lombok.Builder.Default protected double dampingFactor = 100;
  // gradient keys used for ensuring order when getting and setting the gradient
  // @lombok.Builder.Default
  // protected List<String> variables = new ArrayList<>();
  @Getter @Setter @lombok.Builder.Default private boolean miniBatch = false;
  /** A seed for this network, will be random if not specified. */
  @Getter @Setter @lombok.Builder.Default private long seed = new Random().nextLong();
  /**
   * The default {@link CacheMode} for this configuration. Will be set to "NONE" if not specified
   * otherwise. This method defines how/if preOutput cache is handled: NONE: cache disabled (default
   * value) HOST: Host memory will be used DEVICE: GPU memory will be used (on CPU backends effect
   * will be the same as for HOST)
   *
   * <p>Valid values are<br>
   * CacheMode.NONE,<br>
   * CacheMode.HOST or<br>
   * CacheMode.DEVICE<br>
   *
   * @param cacheMode
   */
  @NonNull @Getter @Setter @lombok.Builder.Default private CacheMode cacheMode = CacheMode.NONE;

  /**
   * The name for this configuration. Defaults to "Anonymous INeuralNetworkConfiguration" if it is
   * not specified.
   */
  @lombok.Builder.Default @Getter private String name = "Anonymous INeuralNetworkConfiguration";
  /** The {@link InputType} of the data for this network configuration */
  @Getter @Setter private InputType inputType;
  /**
   * Set the DataType for the network parameters and activations for all layers in the network.
   * Default: Float
   *
   * @param dataType Datatype to use for parameters and activations
   */
  @Getter @Setter @lombok.Builder.Default @NonNull private DataType dataType = DataType.FLOAT;
  /**
   * Whether to override the nIn configuration forcibly upon construction. Default value is true.
   *
   * @return builder pattern
   */
  @Getter @Setter @lombok.Builder.Default private boolean overrideNinUponBuild = true;
  /**
   * Enabled by default. If enabled, the output layer configuration will be validated, to throw an
   * exception on likely invalid outputs - such as softmax + nOut=1, or LossMCXENT + Tanh.<br>
   * If disabled (false) no output layer validation will be performed.<br>
   * Disabling this validation is not recommended, as the configurations that fail validation
   * usually will not be able to learn correctly. However, the option to disable this validation is
   * provided for advanced users when creating non-standard architectures.
   *
   * @param validate If true: validate output layer configuration. False: don't validate
   */
  @Getter @Setter @lombok.Builder.Default private boolean validateOutputLayerConfig = true;
  /**
   * Enabled by default. If enabled, an exception will be throw when using the (invalid) combination
   * of truncated backpropagation through time (TBPTT) with either a GlobalPoolingLayer or
   * LastTimeStepLayer.<br>
   * It is possible to disable this validation to allow what is almost certainly an invalid
   * configuration to be used, however this is not recommended.
   *
   * @param validate Whether TBPTT validation should be performed
   */
  @Getter @Setter @lombok.Builder.Default private boolean validateTbpttConfig = true;
  /**
   * Gradient updater configuration. For example, {@link org.nd4j.linalg.learning.config.Adam} or
   * {@link org.nd4j.linalg.learning.config.Nesterovs}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param updater Updater to use
   */
  @Getter @Setter private IUpdater updater;
  /**
   * Gradient normalization strategy. Used to specify gradient renormalization, gradient clipping
   * etc. See {@link GradientNormalization} for details<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param gradientNormalization Type of normalization to use. Defaults to None.
   * @see GradientNormalization
   */
  @Getter @Setter @NonNull @lombok.Builder.Default
  private GradientNormalization gradientNormalization = GradientNormalization.None;
  /**
   * Threshold for gradient normalization, only used for GradientNormalization.ClipL2PerLayer,
   * GradientNormalization.ClipL2PerParamType, and
   * GradientNormalization.ClipElementWiseAbsoluteValue<br>
   * Not used otherwise.<br>
   * L2 threshold for first two types of clipping, or absolute value threshold for last type of
   * clipping.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   */
  @Getter @Setter private double gradientNormalizationThreshold;
  /**
   * Activation function / neuron non-linearity<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   */
  @Getter @Setter private IActivation activation;
  // whether to constrain the gradient to unit norm or not
  @Getter @Setter private StepFunction stepFunction;

  @Getter @Setter @lombok.Builder.Default
  private OptimizationAlgorithm optimizationAlgo =
      OptimizationAlgorithm.STOCHASTIC_GRADIENT_DESCENT;

  @Getter @Setter @lombok.Builder.Default private int maxNumLineSearchIterations = 5;
  /**
   * Set the regularization for the parameters (excluding biases) - for example {@link WeightDecay}
   * <br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.<br>
   *
   * @param regularization Regularization to apply for the network parameters/weights (excluding
   *     biases)
   */
  @Getter @lombok.Builder.Default private List<Regularization> regularization = new ArrayList<>();
  /**
   * Set the regularization for the biases only - for example {@link WeightDecay}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.<br>
   *
   * @param regularizationBias Regularization to apply for the network biases only
   */
  @Getter @lombok.Builder.Default
  private List<Regularization> regularizationBias = new ArrayList<>();

  @Getter @Setter @lombok.Builder.Default private IUpdater iUpdater = new Sgd();
  /**
   * Gradient updater configuration, for the biases only. If not set, biases will use the updater as
   * set by {@link #setIUpdater(IUpdater)}<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param updater Updater to use for bias parameters
   */
  @Getter @Setter @lombok.Builder.Default private IUpdater biasUpdater = null;

  @Getter @Setter @lombok.Builder.Default
  private IActivation activationFn = new ActivationSigmoid();
  /**
   * Weight initialization scheme to use, for initial weight values Note: values set by this method
   * will be applied to all applicable layers in the network, unless a different value is explicitly
   * set on a given layer. In other words: values set via this method are used as the default value,
   * and can be overridden on a per-layer basis.
   */
  @Getter @Setter @lombok.Builder.Default private IWeightInit weightInit = new WeightInitXavier();
  /**
   * Sets the convolution mode for convolutional layers, which impacts padding and output sizes. See
   * {@link ConvolutionMode} for details. Defaults to ConvolutionMode.TRUNCATE<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param convolutionMode Convolution mode to use
   */
  @Getter @Setter @lombok.Builder.Default
  private ConvolutionMode convolutionMode = ConvolutionMode.Truncate;
  /**
   * Sets the cuDNN algo mode for convolutional layers, which impacts performance and memory usage
   * of cuDNN. See {@link ConvolutionLayer.AlgoMode} for details. Defaults to "PREFER_FASTEST", but
   * "NO_WORKSPACE" uses less memory. <br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param cudnnAlgoMode cuDNN algo mode to use
   */
  @Getter @Setter @lombok.Builder.Default
  private ConvolutionLayer.AlgoMode cudnnAlgoMode = ConvolutionLayer.AlgoMode.PREFER_FASTEST;

  @Getter @Setter @lombok.Builder.Default private boolean minimize = true;
  /**
   * Set the dropout for all layers in this network<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis. * Dropout
   * probability. This is the probability of <it>retaining</it> each input activation value for a
   * layer. * dropOut(x) will keep an input activation with probability x, and set to 0 with
   * probability 1-x.<br>
   * * dropOut(0.0) is a special value / special case - when set to 0.0., dropout is disabled (not
   * applied). Note * that a dropout value of 1.0 is functionally equivalent to no dropout: i.e.,
   * 100% probability of retaining * each input activation.<br>
   * *
   *
   * <p>* Note 1: Dropout is applied at training time only - and is automatically not applied at
   * test time * (for evaluation, etc)<br>
   * * Note 2: This sets the probability per-layer. Care should be taken when setting lower values
   * for * complex networks (too much information may be lost with aggressive (very low) dropout
   * values).<br>
   * * Note 3: Frequently, dropout is not applied to (or, has higher retain probability for) input
   * (first layer) * layers. Dropout is also often not applied to output layers. This needs to be
   * handled MANUALLY by the user * - set .dropout(0) on those layers when using global dropout
   * setting.<br>
   * * Note 4: Implementation detail (most users can ignore): DL4J uses inverted dropout, as
   * described here: * <a
   * href="http://cs231n.github.io/neural-networks-2/">http://cs231n.github.io/neural-networks-2/</a>
   * * *<br>
   * * Note: values set by this method will be applied to all applicable layers in the network,
   * unless a different * value is explicitly set on a given layer. In other words: values set via
   * this method are used as the default * value, and can be overridden on a per-layer basis. *
   * * @param inputRetainProbability Dropout probability (probability of retaining each input
   * activation value for a layer) * @see #dropOut(IDropout)
   *
   * @param dropout Dropout, such as {@link Dropout}, {@link
   *     org.deeplearning4j.nn.conf.dropout.GaussianDropout}, {@link
   *     org.deeplearning4j.nn.conf.dropout.GaussianNoise} etc
   * @return
   */
  @Getter @Setter private IDropout idropOut;
  /**
   * Set the weight noise (such as {@link org.deeplearning4j.nn.conf.weightnoise.DropConnect} and
   * {@link org.deeplearning4j.nn.conf.weightnoise.WeightNoise}) for the layers in this network.<br>
   * Note: values set by this method will be applied to all applicable layers in the network, unless
   * a different value is explicitly set on a given layer. In other words: values set via this
   * method are used as the default value, and can be overridden on a per-layer basis.
   *
   * @param weightNoise Weight noise instance to use
   */
  @Getter @Setter private IWeightNoise weightNoise;

  @Getter @Setter @lombok.Builder.Default private double biasInit = 0.0;
  @Getter @Setter @lombok.Builder.Default private double gainInit = 1.0;

  /**
   * Handle {@link WeightInit} and {@link Distribution} from legacy configs in Json format. Copied
   * from handling of {@link Activation} above.
   *
   * @return True if all is well and layer iteration shall continue. False else-wise.
   */
  private static boolean handleLegacyWeightInitFromJson(
      String json, LayerConfiguration l, ObjectMapper mapper, JsonNode confs, int layerCount) {
    if ((l instanceof BaseLayerConfiguration)
        && ((BaseLayerConfiguration) l).getWeightInit() == null) {
      try {
        JsonNode jsonNode = mapper.readTree(json);
        if (confs == null) {
          confs = jsonNode.get("confs");
        }
        if (confs instanceof ArrayNode) {
          ArrayNode layerConfs = (ArrayNode) confs;
          JsonNode outputLayerNNCNode = layerConfs.get(layerCount);
          if (outputLayerNNCNode == null) {
            return false; // Should never happen...
          }
          JsonNode layerWrapperNode = outputLayerNNCNode.get("layer");

          if (layerWrapperNode == null || layerWrapperNode.size() != 1) {
            return true;
          }

          JsonNode layerNode = layerWrapperNode.elements().next();
          JsonNode weightInit =
              layerNode.get("weightInit"); // Should only have 1 element: "dense", "output", etc
          JsonNode distribution = layerNode.get("dist");

          Distribution dist = null;
          if (distribution != null) {
            dist = mapper.treeToValue(distribution, Distribution.class);
          }

          if (weightInit != null) {
            final IWeightInit wi =
                WeightInit.valueOf(weightInit.asText()).getWeightInitFunction(dist);
            ((BaseLayerConfiguration) l).setWeightInit(wi);
          }
        }

      } catch (IOException e) {
        log.warn(
            "ILayer with null WeightInit detected: " + l.getLayerName() + ", could not parse JSON",
            e);
      }
    }
    return true;
  }

  /**
   * Object mapper for serialization of configurations
   *
   * @return
   */
  public static ObjectMapper mapperYaml() {
    return JsonMappers.getMapperYaml();
  }

  /**
   * Object mapper for serialization of configurations
   *
   * @return
   */
  public static ObjectMapper mapper() {
    return JsonMappers.getMapper();
  }

  public static NeuralNetBaseBuilderConfiguration fromYaml(String input) {
    throw new RuntimeException("Needs fixing - not supported."); // TODO
  }

  /**
   * @return JSON representation of NN configuration
   */
  public String toYaml() {
    ObjectMapper mapper = NeuralNetBaseBuilderConfiguration.mapperYaml();
    synchronized (mapper) {
      try {
        return mapper.writeValueAsString(this);
      } catch (com.fasterxml.jackson.core.JsonProcessingException e) {
        throw new RuntimeException(e);
      }
    }
  }

  /**
   * @return JSON representation of NN configuration
   */
  public String toJson() {
    ObjectMapper mapper = NeuralNetBaseBuilderConfiguration.mapper();
    synchronized (mapper) {
      // JSON mappers are supposed to be thread safe: however, in practice they seem to miss fields
      // occasionally
      // when writeValueAsString is used by multiple threads. This results in invalid JSON. See
      // issue #3243
      try {
        return mapper.writeValueAsString(this);
      } catch (com.fasterxml.jackson.core.JsonProcessingException e) {
        throw new RuntimeException(e);
      }
    }
  }

  @Override
  public NeuralNetBaseBuilderConfiguration clone() {
    NeuralNetBaseBuilderConfiguration clone;
    try {
      clone = (NeuralNetBaseBuilderConfiguration) super.clone();
    } catch (CloneNotSupportedException ex) {
      throw new RuntimeException(ex);
    }
    if (clone.stepFunction != null) {
      clone.stepFunction = clone.stepFunction.clone();
    }
    /** if (clone.variables != null) { clone.variables = new ArrayList<>(clone.variables); } */
    clone.getInnerConfigurations().addAll(innerConfigurations);

    if (clone.getInputPreProcessors() != null) {
      Map<Integer, InputPreProcessor> map = new HashMap<>();
      for (Map.Entry<Integer, InputPreProcessor> entry : clone.getInputPreProcessors().entrySet()) {
        map.put(entry.getKey(), entry.getValue().clone());
      }
      clone.getInputPreProcessors().clear();
      clone.getInputPreProcessors().putAll(map);
    }

    clone.setInferenceWorkspaceMode(this.inferenceWorkspaceMode);
    clone.setTrainingWorkspaceMode(this.trainingWorkspaceMode);
    clone.setCacheMode(this.cacheMode);
    clone.setValidateOutputLayerConfig(this.validateOutputLayerConfig);
    clone.setDataType(this.dataType);

    return clone;
  }

  public abstract static class NeuralNetBaseBuilderConfigurationBuilder<
      C extends NeuralNetBaseBuilderConfiguration,
      B extends NeuralNetBaseBuilderConfiguration.NeuralNetBaseBuilderConfigurationBuilder<C, B>> {

    List<Object> innerConfigurations$value = new ArrayList<>(); // initialize with an empty list

    /**
     * Set constraints to be applied to all layers. Default: no constraints.<br>
     * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
     * regularization, etc). These constraints are applied at each iteration, after the parameters
     * have been updated.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.
     *
     * @param constraints Constraints to apply to all weight parameters of all layers
     */
    public B constrainWeights(LayerConstraint... constraints) {
      contrainWeights$value = Arrays.asList(constraints);
      contrainWeights$set = true;
      return (B) this;
    }

    /**
     * For the (perhaps partially constructed) network configuration, return a list of activation
     * sizes for each layer in the network.<br>
     * Note: To use this method, the network input type must have been set using {@link
     * #setInputType(InputType)} first
     *
     * @return A list of activation types for the network, indexed by layer number
     */
    public List<InputType> getLayerActivationTypes() {
      Preconditions.checkState(
          inputType != null,
          "Can only calculate activation types if input type has"
              + "been set. Use setInputType(InputType)");

      throw new RuntimeException(
          "Error calculating layer activation types: error instantiating MultiLayerConfiguration");
    }

    /**
     * Set constraints to be applied to all layers. Default: no constraints.<br>
     * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
     * regularization, etc). These constraints are applied at each iteration, after the parameters
     * have been updated.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.
     *
     * @param constraints Constraints to apply to all parameters of all layers
     */
    public B constrainAllParameters(LayerConstraint... constraints) {
      allParamContraints$value = Arrays.asList(constraints);
      allParamContraints$set = true;
      return (B) this;
    }

    /**
     * Set constraints to be applied to all layers. Default: no constraints.<br>
     * Constraints can be used to enforce certain conditions (non-negativity of parameters, max-norm
     * regularization, etc). These constraints are applied at each iteration, after the parameters
     * have been updated.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.
     *
     * @param constraints Constraints to apply to all bias parameters of all layers
     */
    public B constrainBias(LayerConstraint... constraints) {
      biasConstraints$value = Arrays.asList(constraints);
      biasConstraints$set = true;
      return (B) this;
    }

    /**
     * Specify the processors. These are used at each layer for doing things like normalization and
     * shaping of input.
     *
     * @param processor what to use to preProcess the data.
     * @return builder pattern
     */
    public B inputPreProcessor(Integer layer, InputPreProcessor processor) {
      inputPreProcessors$value.put(layer, processor);
      inputPreProcessors$set = true;
      return (B) this;
    }

    /**
     * Set layer at index
     *
     * @param index where to insert
     * @param layer the layer
     * @return builder
     */
    public B layer(Integer index, @NonNull LayerConfiguration layer) {
      innerConfigurations$value.add(index, layer);
      innerConfigurations$set = true;
      return (B) this;
    }

    /**
     * Add a layer
     *
     * @param layer the layer
     * @return builder
     */
    public B layer(@NonNull LayerConfiguration layer) {
      innerConfigurations$value.add(layer);
      innerConfigurations$set = true;
      return (B) this;
    }

    // TODO this is a dirty workaround
    public boolean isOverrideNinUponBuild() {
      return isOverrideNinUponBuild();
    }

    /** Specify additional layer configurations */
    @Deprecated
    public B layersFromArray(@NonNull LayerConfiguration[] arrLayers) {
      innerConfigurations$value.addAll(List.of(arrLayers));
      innerConfigurations$set = true;
      return (B) this;
    }

    /** Specify additional layer configurations */
    @Deprecated
    public B layersFromList(@NonNull List<LayerConfiguration> listLayers) {
      innerConfigurations$value.addAll(listLayers);
      innerConfigurations$set = true;
      return (B) this;
    }

    /**
     * L1 regularization coefficient for the weights (excluding biases).<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.
     */
    public B l1(double l1) {
      // Check if existing L1 exists; if so, replace it
      NetworkUtils.removeInstances(regularization$value, L1Regularization.class);
      if (l1 > 0.0) {
        regularization$value.add(new L1Regularization(l1));
      }
      regularization$set = true;
      return (B) this;
    }

    /**
     * L2 regularization coefficient for the weights (excluding biases).<br>
     * <b>Note</b>: Generally, {@link WeightDecay} (set via {@link #weightDecay(double)} should be
     * preferred to L2 regularization. See {@link WeightDecay} javadoc for further details.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.<br>
     * Note: L2 regularization and weight decay usually should not be used together; if any weight
     * decay (or L2) has been added for the biases, these will be removed first.
     *
     * @see #weightDecay(double, boolean)
     */
    public B l2(double l2) {
      // Check if existing L2 exists; if so, replace it. Also remove weight decay - it doesn't make
      // sense to use both
      NetworkUtils.removeInstances(regularization$value, L2Regularization.class);
      if (l2 > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            regularization$value,
            WeightDecay.class,
            "WeightDecay regularization removed: incompatible with added L2 regularization");
        regularization$value.add(new L2Regularization(l2));
      }
      regularization$set = true;
      return (B) this;
    }

    /**
     * L1 regularization coefficient for the bias.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.
     */
    public B l1Bias(double l1Bias) {
      NetworkUtils.removeInstances(regularizationBias$value, L1Regularization.class);
      if (l1Bias > 0.0) {
        regularizationBias$value.add(new L1Regularization(l1Bias));
      }
      regularizationBias$set = true;
      return (B) this;
    }

    /**
     * L2 regularization coefficient for the bias.<br>
     * <b>Note</b>: Generally, {@link WeightDecay} (set via {@link #weightDecayBias(double,
     * boolean)} should be preferred to L2 regularization. See {@link WeightDecay} javadoc for
     * further details.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.<br>
     * Note: L2 regularization and weight decay usually should not be used together; if any weight
     * decay (or L2) has been added for the biases, these will be removed first.
     *
     * @see #weightDecayBias(double, boolean)
     */
    public B l2Bias(double l2Bias) {
      NetworkUtils.removeInstances(regularizationBias$value, L2Regularization.class);
      if (l2Bias > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            regularizationBias$value,
            WeightDecay.class,
            "L2 bias regularization removed: incompatible with added WeightDecay regularization");
        regularizationBias$value.add(new L2Regularization(l2Bias));
      }
      return (B) this;
    }

    /**
     * Add weight decay regularization for the network parameters (excluding biases).<br>
     * This applies weight decay <i>with</i> multiplying the learning rate - see {@link WeightDecay}
     * for more details.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @see #weightDecay(double, boolean)
     */
    public B weightDecay(double coefficient) {
      return weightDecay(coefficient, true);
    }

    /**
     * Add weight decay regularization for the network parameters (excluding biases). See {@link
     * WeightDecay} for more details.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @param applyLR Whether the learning rate should be multiplied in when performing weight decay
     *     updates. See {@link WeightDecay} for more details.
     * @see #weightDecay(double, boolean)
     */
    public B weightDecay(double coefficient, boolean applyLR) {
      // Check if existing weight decay if it exists; if so, replace it. Also remove L2 - it doesn't
      // make sense to use both
      NetworkUtils.removeInstances(regularization$value, WeightDecay.class);
      if (coefficient > 0.0) {
        NetworkUtils.removeInstancesWithWarning(
            regularization$value,
            L2Regularization.class,
            "L2 regularization removed: incompatible with added WeightDecay regularization");
        regularization$value.add(new WeightDecay(coefficient, applyLR));
      }
      regularization$set = true;
      return (B) this;
    }

    /**
     * Weight decay for the biases only - see {@link #weightDecay(double)} for more details. This
     * applies weight decay <i>with</i> multiplying the learning rate.<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     * @see #weightDecayBias(double, boolean)
     */
    public B weightDecayBias(double coefficient) {
      return weightDecayBias(coefficient, true);
    }

    /**
     * Weight decay for the biases only - see {@link #weightDecay(double)} for more details<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.<br>
     *
     * @param coefficient Weight decay regularization coefficient
     */
    public B weightDecayBias(double coefficient, boolean applyLR) {
      // Check if existing weight decay if it exists; if so, replace it. Also remove L2 - it doesn't
      // make sense to use both
      NetworkUtils.removeInstances(regularizationBias$value, WeightDecay.class);
      if (coefficient > 0) {
        NetworkUtils.removeInstancesWithWarning(
            regularizationBias$value,
            L2Regularization.class,
            "L2 bias regularization removed: incompatible with added WeightDecay regularization");
        regularizationBias$value.add(new WeightDecay(coefficient, applyLR));
      }
      regularization$set = true;
      return (B) this;
    }

    /**
     * Activation function / neuron non-linearity<br>
     * Note: values set by this method will be applied to all applicable layers in the network,
     * unless a different value is explicitly set on a given layer. In other words: values set via
     * this method are used as the default value, and can be overridden on a per-layer basis.
     */
    @Deprecated
    public B activation(@NonNull Activation activation) {
      return (B) activationFn(activation.getActivationFunction());
    }

    /**
     * legacy code, does nothing
     *
     * @return
     */
    @Deprecated
    public B list() {
      return (B) this;
    }

    /**
     * Set weight initialization scheme to random sampling via the specified distribution.
     * Equivalent to: {@code .weightInit(new WeightInitDistribution(distribution))} Note: values set
     * by this method will be applied to all applicable layers in the network, unless a different
     * value is explicitly set on a given layer. In other words: values set via this method are used
     * as the default value, and can be overridden on a per-layer basis.
     *
     * @param distribution Distribution to use for weight initialization
     */
    public B weightInit(@NonNull Distribution distribution) {
      this.weightInit$value = new WeightInitDistribution(distribution);
      this.weightInit$set = true;
      return (B) this;
    }

    public B weightInit(@NonNull WeightInit weightInit) {
      this.weightInit$value = weightInit.getWeightInitFunction();
      this.weightInit$set = true;
      return (B) this;
    }

    public B weightInit(@NonNull IWeightInit iWeightInit) {
      this.weightInit$value = iWeightInit;
      this.weightInit$set = true;
      return (B) this;
    }

    /**
     * Same as {@link #weightInit(Distribution)}.
     *
     * @param distribution
     * @return
     */
    public B dist(@NonNull Distribution distribution) {
      return (B) weightInit(distribution);
    }

    public B dropOut(@NonNull IDropout dropout) {
      return (B) idropOut(dropout);
    }

    /**
     * Creates a new {@link Dropout} and sets the dropout in the builder for this configuration
     *
     * @param dropout activationRetainProbability
     * @return builder
     */
    public B dropOut(double dropout) {
      return (B) idropOut(new Dropout(dropout));
    }

    /**
     * Add multiple inner neural net configurations at once
     *
     * @param confs list of configurations
     * @return builder
     */
    @Deprecated
    public B confs(@NonNull List<NeuralNetConfiguration> confs) {
      innerConfigurations$value.addAll(confs);
      innerConfigurations$set = true;
      return (B) this;
    }
  }
}
