/*
 *  ******************************************************************************
 *  *
 *  *
 *  * This program and the accompanying materials are made available under the
 *  * terms of the Apache License, Version 2.0 which is available at
 *  * https://www.apache.org/licenses/LICENSE-2.0.
 *  *
 *  *  See the NOTICE file distributed with this work for additional
 *  *  information regarding copyright ownership.
 *  * Unless required by applicable law or agreed to in writing, software
 *  * distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
 *  * WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
 *  * License for the specific language governing permissions and limitations
 *  * under the License.
 *  *
 *  * SPDX-License-Identifier: Apache-2.0
 *  *****************************************************************************
 */

package org.deeplearning4j.nn.layers.wrapper;

import java.util.Collection;
import java.util.Map;
import lombok.Data;
import lombok.NonNull;
import net.brutex.ai.dnn.api.IModel;
import org.deeplearning4j.nn.api.Layer;
import org.deeplearning4j.nn.api.MaskState;
import org.deeplearning4j.nn.api.ITraininableLayerConfiguration;
import org.deeplearning4j.nn.conf.CacheMode;
import org.deeplearning4j.nn.conf.NeuralNetConfiguration;
import org.deeplearning4j.nn.conf.layers.BaseLayerConfiguration;
import org.deeplearning4j.nn.conf.layers.LayerConfiguration;
import org.deeplearning4j.nn.gradient.Gradient;
import org.deeplearning4j.nn.layers.AbstractLayer;
import org.deeplearning4j.nn.layers.LayerHelper;
import org.deeplearning4j.nn.workspace.LayerWorkspaceMgr;
import org.deeplearning4j.optimize.api.ConvexOptimizer;
import org.deeplearning4j.optimize.api.TrainingListener;
import org.jetbrains.annotations.NotNull;
import org.nd4j.common.primitives.Pair;
import org.nd4j.linalg.api.ndarray.INDArray;
import org.nd4j.linalg.dataset.api.DataSet;
import org.nd4j.linalg.dataset.api.MultiDataSet;
import org.nd4j.linalg.dataset.api.iterator.DataSetIterator;
import org.nd4j.linalg.dataset.api.iterator.MultiDataSetIterator;

@Data
public abstract class BaseWrapperLayer extends AbstractLayer {

  protected Layer underlying;

  public BaseWrapperLayer(@NonNull Layer underlying) {
    this.underlying = underlying;
    this.setLayerConfiguration( underlying.getLayerConfiguration() );
  }

  @Override
  public BaseLayerConfiguration getTypedLayerConfiguration() {
    return (BaseLayerConfiguration) underlying.getLayerConfiguration();
  }

  /**
   * This method returns updater state (if applicable), null otherwise
   *
   * @return
   */
  @Override
  public INDArray updaterState() {
    return underlying.updaterState();
  }

  /**
   * This method fits model with a given DataSet
   *
   * @param dataSet
   */
  @Override
  public void fit(DataSet dataSet) {
underlying.fit(dataSet);
  }

  /**
   * This method fits model with a given MultiDataSet
   *
   * @param dataSet
   */
  @Override
  public void fit(MultiDataSet dataSet) {
underlying.fit(dataSet);
  }

  /**
   * This method fits model with a given DataSetIterator
   *
   * @param iterator
   */
  @Override
  public void fit(DataSetIterator iterator) {
underlying.fit(iterator);
  }

  /**
   * This method fits model with a given MultiDataSetIterator
   *
   * @param iterator
   */
  @Override
  public void fit(MultiDataSetIterator iterator) {
underlying.fit(iterator);
  }

  /**
   * @param netConfiguration
   */
  @Override
  public void setNetConfiguration(@NonNull NeuralNetConfiguration netConfiguration) {
underlying.setNetConfiguration(netConfiguration);
  }


  /**
   * Get a reference to the network this layer is part of.
   *
   * @return
   */
  @Override
  public IModel getNet() {
    return underlying.getNet();
  }

  /**
   * @return 1d parameter vector
   */
  @Override
  public INDArray getParams() {
    return underlying.getParams();
  }

  /**
   * Return the configuration of this layer (which is the configuration of the underlying layer in
   * this case
   *
   * @return the underlying layer configuration
   */
  @Override
  public LayerConfiguration getLayerConfiguration() {
    return underlying.getLayerConfiguration();
  }

  @Override
  public void setLayerConfiguration(LayerConfiguration layerConfiguration) {
    underlying.setLayerConfiguration(layerConfiguration);
  }

  @Override
  public void setCacheMode(CacheMode mode) {
    underlying.setCacheMode(mode);
  }

  @Override
  public double calcRegularizationScore(boolean backpropParamsOnly) {
    return underlying.calcRegularizationScore(backpropParamsOnly);
  }

  @Override
  public Type type() {
    return underlying.type();
  }

  @Override
  public Pair<Gradient, INDArray> backpropGradient(INDArray epsilon,
      LayerWorkspaceMgr workspaceMgr) {
    return underlying.backpropGradient(epsilon, workspaceMgr);
  }

  @Override
  public INDArray activate(boolean training, LayerWorkspaceMgr workspaceMgr) {
    return underlying.activate(training, workspaceMgr);
  }

  @Override
  public INDArray activate(INDArray input, boolean training, LayerWorkspaceMgr workspaceMgr) {
    return underlying.activate(input, training, workspaceMgr);
  }

  @NotNull
  @Override
  public Collection<TrainingListener> getTrainingListeners() {
    return underlying.getTrainingListeners();
  }

  @Override
  public void addTrainingListeners(TrainingListener... listeners) {
    underlying.addTrainingListeners(listeners);
  }

  @Override
  public void fit() {
    underlying.fit();
  }

  @Override
  public void update(Gradient gradient) {
    underlying.update(gradient);
  }

  @Override
  public void update(INDArray gradient, String paramType) {
    underlying.update(gradient, paramType);
  }

  @Override
  public double getScore() {
    return underlying.getScore();
  }

  @Override
  public void computeGradientAndScore(LayerWorkspaceMgr workspaceMgr) {
    underlying.computeGradientAndScore(workspaceMgr);
  }

  @Override
  public INDArray getModelParams() {
    return underlying.getParams();
  }

  @Override
  public long numParams() {
    return underlying.numParams();
  }

  @Override
  public long numParams(boolean backwards) {
    return underlying.numParams();
  }

  @Override
  public void setParams(INDArray params) {
    underlying.setParams(params);
  }

  @Override
  public void setParamsViewArray(INDArray params) {
    underlying.setParamsViewArray(params);
  }

  @Override
  public INDArray getGradientsViewArray() {
    return underlying.getGradientsViewArray();
  }

  @Override
  public void setBackpropGradientsViewArray(INDArray gradients) {
    underlying.setBackpropGradientsViewArray(gradients);
  }

  @Override
  public void fit(INDArray data, LayerWorkspaceMgr workspaceMgr) {
    underlying.fit(data, workspaceMgr);
  }

  @Override
  public Gradient gradient() {
    return underlying.gradient();
  }

  @Override
  public Pair<Gradient, Double> gradientAndScore() {
    return underlying.gradientAndScore();
  }

  @Override
  public int batchSize() {
    return underlying.batchSize();
  }

  @Override
  public NeuralNetConfiguration getNetConfiguration() {
    return underlying.getNetConfiguration();
  }

  @Override
  public INDArray input() {
    return underlying.input();
  }

  @Override
  public ConvexOptimizer getOptimizer() {
    return underlying.getOptimizer();
  }

  @Override
  public INDArray getParam(String param) {
    return underlying.getParam(param);
  }

  @Override
  public Map<String, INDArray> getParamTable() {
    return underlying.getParamTable();
  }

  /**
   * Setter for the param table
   *
   * @param paramTable Map&lt;String, INDArray>
   */
  @Override
  public void setParamTable(Map paramTable) {
    underlying.setParamTable(paramTable);
  }

  @Override
  public Map<String, INDArray> getParamTable(boolean backpropParamsOnly) {
    return underlying.getParamTable(backpropParamsOnly);
  }

  @Override
  public void setParam(String key, INDArray val) {
    underlying.setParam(key, val);
  }

  @Override
  public void clear() {
    underlying.clear();
  }

  @Override
  public void applyConstraints(int iteration, int epoch) {
    underlying.applyConstraints(iteration, epoch);
  }

  @Override
  public void init() {
    underlying.init();
  }

  @Override
  public int getIndex() {
    return underlying.getIndex();
  }

  @Override
  public void setIndex(int index) {
    underlying.setIndex(index);
  }

  @Override
  public int getIterationCount() {
    return underlying.getIterationCount();
  }

  @Override
  public void setIterationCount(int iterationCount) {
    underlying.setIterationCount(iterationCount);
  }

  @Override
  public int getEpochCount() {
    return underlying.getEpochCount();
  }

  @Override
  public void setEpochCount(int epochCount) {
    underlying.setEpochCount(epochCount);
  }

  @Override
  public void setInput(INDArray input, LayerWorkspaceMgr workspaceMgr) {
    underlying.setInput(input, workspaceMgr);
  }

  @Override
  public int getInputMiniBatchSize() {
    return underlying.getInputMiniBatchSize();
  }

  @Override
  public void setInputMiniBatchSize(int size) {
    underlying.setInputMiniBatchSize(size);
  }

  @Override
  public INDArray getMaskArray() {
    return underlying.getMaskArray();
  }

  @Override
  public void setMaskArray(INDArray maskArray) {
    underlying.setMaskArray(maskArray);
  }

  @Override
  public boolean isPretrainLayer() {
    return underlying.isPretrainLayer();
  }

  @Override
  public void clearNoiseWeightParams() {
    underlying.clearNoiseWeightParams();
  }

  @Override
  public Pair<INDArray, MaskState> feedForwardMaskArray(INDArray maskArray,
      MaskState currentMaskState, int minibatchSize) {
    return underlying.feedForwardMaskArray(maskArray, currentMaskState, minibatchSize);
  }

  @Override
  public void allowInputModification(boolean allow) {
    underlying.allowInputModification(allow);
  }

  @Override
  public LayerHelper getHelper() {
    return underlying.getHelper();
  }

  @Override
  public ITraininableLayerConfiguration getTrainingConfig() {
    return underlying.getTrainingConfig();
  }

  @Override
  public boolean updaterDivideByMinibatch(String paramName) {
    return underlying.updaterDivideByMinibatch(paramName);
  }

  @Override
  public void close() {
    //No-op for individual layers
  }
}
